/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                                field.h
 *
 *  Definitions of the main FieldObj concentration field class (derived from
 *  the VectorObj class), the daughter BufferObj buffer concentration field 
 *  class, and the compound buffer array class BufferArray.
 *
 *  FieldObj methods include all low-level ADI difference schemes routines.
 *  Highest-level numeric engine routines are defined in "calc.cpp"       
 *       
 *  FieldObj and BufferArray variables are member of the main SimulationObj
 *  object, defined in simulation.h / simulation.cpp
 *
 ******************************************************************************/

#ifndef CALC_FIELD_H_included
#define CALC_FIELD_H_included

#define SOURCE_EPS  1.0e-4
#define MIN_CURRENT 1.0e-12

const int  VARY_MASK = 127;
const int  DIRICHLET = 128;
const int  REGION_MASK = 255;  // DIRICHLET + VARY_MASK
const int  _OUTSIDE_ = 256;

const int  BC_ID_MASK = 127;  // 5 bits for bc type id - at most 32 different surface types
const int  XSHIFT = 9;
const int  YSHIFT = 16;
const int  ZSHIFT = 23;


typedef int (*FieldMethod)(class FieldObj &, VectorObj &, class BufferArray &, double dt, double);
typedef int (*BufferMethod)(class BufferObj &, VectorObj &, class FieldObj &, double dt, double); 
typedef void (FieldObj::*nabla)(long, int, long, double &, double &, double &, double &);

double gaussian(double, double, double, int);
double square(double, double, double, int);

extern int GEOMETRY;

double calcium_gain( FieldObj *Ca, BufferArray *Bufs);


//****************************************************************************
//*                         C L A S S   F I E L D
//****************************************************************************

class FieldObj: public VectorObj
{ 

protected:

  //static  double (*integrate)(double, double, double, int);

  static  double *dxplus, *dxminus, *dxinv, *dx2inv;
  static  double *dyplus, *dyminus, *dyinv, *dy2inv;
  static  double *dzplus, *dzminus, *dzinv, *dz2inv;

  static  double *right, *diag, *sup, *sub;
  static  double K_PUMP;

  void    tridiag(int n);

  int     *source_x0,    *source_nx;
  int     *source_y0,    *source_ny;
  int     *source_z0,    *source_nz;
  double  **source_xsum, **source_ysum, **source_zsum, *weights;

public:

  static  double *bc_deriv, *bc_coef, *bc_lin, *bc_pump;
  static  char   **bc_id;
  static  int    bc_type_num;

  static  int   xsize, ysize, zsize;
  static  long  Size, xysize;

  static  double *xgrid, *ygrid, *zgrid;
  static  double *xcoord, *ycoord, *zcoord;

  static class RegionObj  *Region;
  static class GridObj    *Grid;
  static class BCarrayObj *BCarray;

  static RegionObj *get_region() { return Region; }
  static void setStaticData(RegionObj &r, GridObj &GO, BCarrayObj &BCA);

  static void    init_tridiag(int);
  static void    kill_tridiag();

  static nabla nablaX;
  static nabla nablaY;
  static nabla nablaZ;


  double  *Diff;
  bool    tortDefined; // true if tortuosity is defined 

  double  D;
  int     source_num;

  long    *ptype;
  int     *bccond;

  class   ExpressionObj  *current;
  double  ICa;

  double  bgr; // initialAve;
  char    *ID;
  double  Time;

  void allocate(const char *id = "Field");

  FieldObj(int=0, double=1.0, double tot=0.0, const char *id = "Ca");
  FieldObj(TokenString &, const char *id = "Ca");
  FieldObj(const FieldObj &);

  ~FieldObj();

  FieldObj& operator=(const FieldObj &f);
  FieldObj& operator=(const VectorObj &f);
  FieldObj& operator=(double (*)(double, double, double));
  FieldObj& operator=(const double);

  void init_boundary();

  void set_source(int, double, double, double, double, double, double);
  void set_source_xyz(int id, double c0, long i0, int ix, double sigma, 
                      double *coord, double *grid, int minMask, int maxMask, int incr, int N,
                      int *source_xyz0, int *source_nxyz, double **source_xyzsum);

  void add_sources(VectorObj &, double);
  void add_sources(double dt) { add_sources(*this, dt); }

  void set_bc(int boxid, int, int, int, int, int, int);
  void set_bc(int boxid, int);
  void set_bc(int boxid, const char *, const char *, const char *,
                         const char *, const char *, const char *);
  void set_bc(int boxid, const char *);
  int  get_bc(int boxid, int surface) { return bccond[boxid*6 + surface]; }  

  double get_value(double, double, double);

  long *get_ptype() { return ptype; }

  double getD() { return D; }
  void   setD(double d) { D = d; }

  void nabla1Dx(long i,   int ix, long bcvar, double &, double &, double &, double &);
  void nabla1Dy(long i,   int iy, long bcvar, double &, double &, double &, double &);
  void nabla1Dz(long i,   int iz, long bcvar, double &, double &, double &, double &);
  void nablaSphereR(long i,   int ix, long bcvar, double &, double &, double &, double &);
  void nablaTheta(long i, int iy, long bcvar, double &, double &, double &, double &);
  void nablaDiscR(long i,   int ix, long bcvar, double &, double &, double &, double &);

  void Run1D(double, double, double *, double *);
	     //               void (FieldObj::*)(long, int, long, double &, double &, double &, double &));

  void Run2Dx(double, double, double, double *, double *);
  void Run2Dy(double, double, double, double *, double *);

  void Run3Dx(double, double, double, double, double *, double *);
  void Run3Dy(double, double, double, double, double *, double *);
  void Run3Dz(double, double, double, double, double *, double *);

  int Step(FieldMethod Method, VectorObj &Next, class BufferArray& Bufs, double dt, double T)
    { return (*Method)(*this, Next, Bufs, dt, T); }

  void cleanBoundaries(); 
  void print_ptype(long i);
  void print_ptype();

  double interpolate(double x, double y, double z);

  //  void init(double T=0.0) { Time = T; *this = total; };

  double average1D(double * = 0);
  double average2D(double * = 0);
  double average3D(double * = 0);
  double averageSpherical(double * = 0);
  double averageConical(double * = 0);
  double averageDisc(double * = 0);
  double averageCylindrical(double * = 0);

  double average(double *V = 0)  { 
     switch(GEOMETRY) { case SPHERICAL:    return averageSpherical(V);
                        case CONICAL:      return averageConical(V);
                        case DISC:         return averageDisc(V);
                        case CYLINDRICAL:  return averageCylindrical(V);
                        case CARTESIAN1D:  return average1D(V);
                        case CARTESIAN2D:  return average2D(V);
                        default: return average3D(V); } }
  double gain()  { 
    double V;
    double c = average(&V) - bgr; // - initialAve;
    return c * V;
  }    
 
  double sum();
  double L2norm();
  void   print3d(int n=3);
  friend void print3d(long *);
};


//*****************************************************************************
//                         B U F F E R   C L A S S
//*****************************************************************************

class BufferObj : public FieldObj
{
 protected:
  
 public:

  VectorObj *total;

  double kplus;
  double kminus;
  double KD;

  ~BufferObj() { delete total; };

  BufferObj(TokenString &, const char *);

  BufferObj(const BufferObj &b) : FieldObj(b)
    { kminus = b.kminus; kplus = b.kplus; KD = b.KD; 
      total = new VectorObj(*(b.total)); }

  BufferObj& operator=(const BufferObj &b)
    { 
    *(FieldObj *)this = b; 
    *total = *(b.total);
    kplus = b.kplus; kminus = b.kminus; KD = b.KD; 
    return *this; 
    }

  BufferObj& operator=(const FieldObj &f)
    { *(FieldObj *)this = f; return *this; }

  BufferObj& operator=(const VectorObj &v)
    { *(VectorObj *)this = v; return *this; }

  BufferObj& operator=(double (*f)(double, double, double))
    { *(FieldObj *)this = f; return *this; }

  BufferObj& operator=(const double val)
    { *(VectorObj *)this = val; return *this; }

  void setMembraneBound(double Total);
  void setMembraneLayer(double Total, double depth);

  int Step(BufferMethod Method, VectorObj &Next, FieldObj &Ca, double dt, double T)
    { return (*Method)(*this, Next, Ca, dt, T); }
};

//****************************************************************************
//*                   C L A S S   F I E L D   A R R A Y
//****************************************************************************

class BufferArray
{
private:

public:

  char *ID;
  int buf_num;
  BufferObj **array;

  BufferArray() { array = 0; ID = 0; buf_num = 0; }
  BufferArray(TokenString &params, const char *group="buffer");
  BufferArray(const BufferArray &bufs)  { allocateCopy(bufs); }

  ~BufferArray()  { destroy(); }
 
  void operator=(const BufferArray &bufs) {  // beware: group ID ignored
    if ( (buf_num = bufs.buf_num) > 0 )
      for (int i = 0; i < buf_num; i++) *array[i] = *bufs.array[i];
  }

  void allocateCopy(const BufferArray&);

  void destroy()  {  if (buf_num) {
                       delete [] ID; 
                       for (int i = 0; i < buf_num; i++) delete array[i]; 
                       delete [] array;
		     }
                   };

  void Step(BufferMethod Method, VectorObj &Next, FieldObj& Ca, double dt, double T) { 
    for (int i = 0; i < buf_num; i++) array[i]->Step(Method, Next, Ca, dt, T); }

  double gain() {
    double bound = 0.0;
    for (int i = 0; i < buf_num; i++) bound += array[i]->gain();
  return bound;
  }
    
};

//****************************************************************************


#endif
